"""Module for builtin RDF parser

All builtin [RDF parser][rdflib.parser.Parser] available
in [RDFLib Graphs method `parse`][rdflib.graph.Graph.parse] are listed here.

Registration of builtin and external plugins
is described in [`rdflib.plugin`][rdflib.plugin].
"""
