"""Module for builtin RDF stores.

All builtin [RDF Stores][rdflib.store.Store] available
during [RDFLib Graphs creation][rdflib.graph.Graph] are listed here.

Registration of builtin and external plugins
is described in [`rdflib.plugin`][rdflib.plugin].
"""
