"""Module for builtin RDF Serializer.

All builtin [RDF serializer][rdflib.serializer.Serializer] available
in [RDFLib Graphs method `serialize`][rdflib.graph.Graph.serialize] are listed here.

Registration of builtin and external plugins
is described in [`rdflib.plugin`][rdflib.plugin].
"""
